// A materialmap converter from BF1942 to BF2 ( 8 Bit raw)
// input data is 8 bit RAW or 8/24 bit BMP
// output data is 8 bit RAW
// if an external 'conversiontable.hex' file is provided, it will use that instead of the built-in table
/************************************  includes              ************************/

#include <stdlib.h>
#include <stdio.h>
#include <stdarg.h>
#include <string.h>
#include <ctype.h>

#ifdef DEBUG
  #include <dmalloc.h>
#endif


#ifdef IS_UNIX
  #include <unistd.h>
#else
  #include <direct.h>
#endif

unsigned char Translate[255] = {0,1,6,5,8,8,9,7,3,3,2,2,12,2,8,4,0,0,0,0,0,0 };

/************************************  macros                 ***********************/


#define MAJOR_VERSION 0
#define MINOR_VERSION 1
#define SUB_VERSION "a"

#define true  1
#define false 0
#define TRUE  1
#define FALSE 0

#ifdef IS_UNIX
  #define COMSEP '-'
  #define PATHSEP '/'
  #define LINEFEED "\n"
  #define COMMENTTAG "#"
#else
  #define COMSEP '/'
  #define PATHSEP '\\'
  #define LINEFEED "\r\n"
  #define COMMENTTAG ";"
#endif

#define READ_MODE  "rb"
#define WRITE_MODE "wb"
#define BI_RGB 0
#define BI_RLE8 1
#define BI_RLE4 2
#define BI_BITFIELDS 3

#define DIB_SIGNATURE 0x4D42 /* "BM" */
#define DIB_PALVERSION 0x0300
#define DIB_PELSPERMETER 0xec4  /* plays better than 0xec3 */
#pragma pack (1)
#define PUT_2B(array,offset,value)  \
	(array[offset]   = (char) ((value) & 0xFF), \
	 array[offset+1] = (char) (((value) >> 8) & 0xFF))
#define PUT_4B(array,offset,value)  \
	(array[offset]   = (char) ((value) & 0xFF), \
	 array[offset+1] = (char) (((value) >> 8) & 0xFF), \
	 array[offset+2] = (char) (((value) >> 16) & 0xFF), \
	 array[offset+3] = (char) (((value) >> 24) & 0xFF))
#pragma pack ()
#define FILE_RAW_EXT        ".raw"
#define FILE_BMP_EXT        ".bmp"
#define BUF_SIZE 4096 * 4096 * 4
typedef enum _mapRes
  {
    SM_MAP = 256,
    MD_MAP = 512,
    LG_MAP = 1024,
    XL_MAP = 2048
  } mapRes;
typedef enum _outmapRes
  {
    SM_OUTMAP = 257,
    MD_OUTMAP = 513,
    LG_OUTMAP = 1025,
    XL_OUTMAP = 2049
  } outmapRes;
typedef enum _errFlag
  {
    EF_NONE = 0,
    EF_USAGE,
    EF_NO_JOBS,
    EF_MALLOC,
    EF_FILE_OPEN,
    EF_FILE_SIZE,
    EF_FILE_SEEK,
    EF_FILE_READ,
    EF_FILE_WRITE,
    EF_FILE_NAME,
    EF_NOT_FOUND,
    EF_DATA_MISSING,
    EF_INFO_MISSING,
    EF_MEM_OVERRUN,
    EF_LOOP_TIMEOUT,
    EF_PASSED_NULL,
    EF_BAD_FILE,
    EF_BAD_DATA,
    EF_NOT_SUPPORTED,
    EF_FUNCTION_ERR,
    EF_UNKNOWN_ERROR,
  } errFlag;
#pragma pack (1)
typedef struct _bidHeader
{
  unsigned short sig;
  unsigned int   fileSize;
  unsigned short res1;
  unsigned short res2;
  unsigned int   offBits;
} bidHeader;/*  //__attribute__  ((packed)) bidHeader; */

typedef struct _bidInfoHeader
{
  unsigned int   biSize;
  int            biWidth;
  int            biHeight;
  unsigned short biPlanes;
  unsigned short biBitCount;
  unsigned int   biCompression;
  unsigned int   biSizeImage;
  int            biXPelsPerMeter;
  int            biYPelsPerMeter;
  unsigned int   biClrUsed;
  unsigned int   biClrImportant;
  
} bidInfoHeader;
typedef struct _rgbTriplet 
{
  unsigned char rgbBlue;
  unsigned char rgbGreen;
  unsigned char rgbRed;
} rgbTriplet; /* __attribute__ ((packed)) rgbQuad; */
typedef struct _rgbQuad 
{
  unsigned char rgbBlue;
  unsigned char rgbGreen;
  unsigned char rgbRed;
  unsigned char rgbReserved;
} rgbQuad; /* __attribute__ ((packed)) rgbQuad; */
#pragma pack ()

/************************************  prototypes             ***********************/
int findLn2 ( int p );
int testfill(int length);
void loadBmpHeader ( FILE *fp );
int process(void);
void     parseArgs    ( int argc, char *argv[] );
void     usage        ( const char *name );
void myshutdown ( int err , char *string);
char * dupString ( char *str );
char * fileName ( char *path );
char * strToUpper ( char *str );
int isFile ( char *str );
int isDir ( char *str );
int readinfile(char *infile); 
int readintable(char *infile); 
int outputfile(char *outfile); 
int fileSize ( FILE  *fp );


/* globals */
  char                   *inpath;
  char                   *tabpath;
  char                   *outpath;
  unsigned char indata[BUF_SIZE]; /*data */  
  unsigned char outdata[BUF_SIZE]; /*data */  
  int inputbufSize;
  int outputbufSize;
  int depth; /*color depth of input BMP */ 
  int res; /* map size ; heightmaps have mapsize DIV 4 coordinates */
  int isBmp ;
/* now functions and utilities */ 

int findLn2 ( int p )
{
  int i = 0;
  while (( p >>= 1 ) > 0 ) i++;
  return i;

} /* end findLn2 */


void
loadBmpHeader ( FILE *fp )
{
  bidHeader      header;
  bidInfoHeader  infoHeader;
  int numColors;
  rgbQuad color;

    /* read BMP file header */
  if ( !fread ( &header, sizeof ( bidHeader ), 1, fp ))
      myshutdown ( EF_FILE_READ,
		 "Error reading %s bitmap header\n");

  if ( header.sig != DIB_SIGNATURE )
    myshutdown ( EF_BAD_FILE,
	       "input Bitmap file has bad signature\n");
/*
  if ( (int) header.fileSize != fileSize ( map ) )
    myshutdown ( EF_BAD_FILE,
	       "input bitmap file size does not match header filesize\n");
*/
  if ( !fread ( &infoHeader, sizeof ( bidInfoHeader ), 1, fp ))
      myshutdown ( EF_FILE_READ,
		 "Error reading input file bitmap info header\n");

  if ( infoHeader.biSize != 40 )
    myshutdown ( EF_BAD_DATA,
	       "Unexpected bitmap info header size for input file.\n" );

  if ( infoHeader.biWidth != infoHeader.biHeight )
    myshutdown ( EF_BAD_DATA,
	       "input bitmap file must be square\n" );

  if ( infoHeader.biCompression )
    myshutdown ( EF_BAD_DATA,
	       "input bitmap file compression is unsupported.\n");

  numColors = ( header.offBits - sizeof ( bidHeader ) - sizeof ( bidInfoHeader )) / 4;

  if ( infoHeader.biClrUsed ) {
    if ( numColors != (int) infoHeader.biClrUsed )
      myshutdown ( EF_BAD_DATA,"%s bitmap file allocated colors mismatch.\n");
  } else 
    infoHeader.biClrUsed = numColors;

  depth = findLn2 ( numColors );

  if ( !fread ( &color, sizeof ( rgbQuad ), 1, fp ))
    myshutdown ( EF_BAD_DATA, "Error reading color table in image file.\n");
//  if ( color.rgbRed ) map->io.inverted = TRUE;

  /* set file pointer to data */
  fseek ( fp, header.offBits, SEEK_SET );

  if ( infoHeader.biBitCount == 1 )
    myshutdown ( EF_BAD_DATA,
	       "%s bitmap file must be 8 or 24 bit\n");

  depth = infoHeader.biBitCount;
  res = infoHeader.biWidth;
  printf("Found a BMP with color depth of %d and a width of %d\n", depth,res);
	isBmp = TRUE;
} /* end loadBmpHeader */


int testfill(int length) 
{
int x,d = 0;
int delta = 2;
unsigned char y;

for (x=0;x<length; x += 2 ) {
	y = d / 256;
	outdata[x+1] = 40 + y;
	y = d - (d * 256) >> 8;
	outdata[x] = y;
	d = d + delta;
	if ((d > 32768) | ( d < 0)) delta = -delta;	
	}
	return x;
}
/* here the main code translation takes place : 
   read the input file , get the color picture and output
   truncated bytes according to desired format */ 

int process() {
unsigned char r,g,b;
int x,y,step;
int out = 0;
	step = depth / 8;
for (x=0;x < inputbufSize; x += step) {
		r=indata[x];
		g=indata[x+1];
		b=indata[x+2];
	switch (depth) {
	case 8 : outdata[out] = Translate[r]; /* assemble the value */
			 out++;
			 break; 
	default:                  /* 24 bit picture */
		 	outdata[out] = 0;
		 	out++;
		 	if (isBmp) outdata[out] = g; /* pick the green channel */
		 	else {						/* this format is used by MICRODEM with color table  */
		 	y = g + (256 * b);
		 	if (y > 255) y = 255;
		 	outdata[out] = y >> 8 ;
		 	}
		 	out++;			
			break;
	};
	}
	return out;
}
void
myshutdown ( err ,string)
{
	printf("%s\n",string);
    exit (err);
} /* end myshutdown */


char *
dupString ( char *str )
{
  char *p;

  if ( !( p = (char *) calloc ( strlen(str) + 1,1 )))
    myshutdown ( EF_MALLOC, "Error duplicating string.\n" );
  strncpy ( p, str, strlen(str) );
  return p;

}

char *
fileName ( char *path )
{
  char *p;
  if (( p = strrchr ( path, PATHSEP )) == NULL )
    p = path;
  else p++;
  return p;
}

char *
strToUpper ( char *str )
{
  int i;
  for ( i = 0; i < (int) strlen ( str ); i++ ) str[i] = toupper ( str[i] );
  return str;
}


int
isFile ( char *str )
{
  FILE *fp;

  if ( !( fp = fopen ( str, READ_MODE ))) return FALSE;
  else fclose ( fp );

  return TRUE;
}

int
fileSize ( FILE  *fp )
{
  int filesize;
  int curPointer;

  if ( !fp )
    return -1;
  /* save current file pointer */
  curPointer = ftell ( fp );

    /* set pointer to file end */
  if ( fseek ( fp, 0, SEEK_END ) != 0 )
    return -1;

  /* how far is that exactly? */
  filesize = ftell ( fp );
  /* reset the file pointer */
  fseek ( fp, curPointer, SEEK_SET );
  /* return file size */
  return filesize;
}
int readinfile(char *infile) 
{
  FILE *fp;
  char *s ="";
   
  if ( !( fp = fopen ( infile, READ_MODE ))) 
  	myshutdown ( EF_FILE_READ, "Error opening input file.\n" ); /* no such file */
  if ( !( s = (char *) calloc ( strlen(infile) + 1,1 )))
    myshutdown ( EF_MALLOC, "Error duplicating string.\n" );
 	strncpy(s, strToUpper(infile),strlen(infile));
/* see if we have a BMP or RAW file */
  	if (s[strlen(s)-1] == 'P' ) {
/* this routine leaves the filepointer at the start of data */
  		loadBmpHeader(fp); 
   	 	inputbufSize = fileSize (fp) - (ftell (fp));
   	 } else {
   	 inputbufSize = fileSize (fp);
   	 depth=8;
   	 isBmp = FALSE;
   	 } 
    printf("filesize is %d\n", inputbufSize);
 	// indata = malloc ( bufSize );
      if ( !fread ( indata, inputbufSize , 1, fp ))
      myshutdown ( EF_FILE_READ, "Error reading from input file.\n" );
      
	fclose(fp);
	return TRUE;
 } 

int readintable(char *infile) 
{
  FILE *fp;
  char *s ="";
   
  if ( !( fp = fopen ( infile, READ_MODE ))) 
  	myshutdown ( EF_FILE_READ, "Error opening table file.\n" ); /* no such file */
  if ( !( s = (char *) calloc ( strlen(infile) + 1,1 )))
    myshutdown ( EF_MALLOC, "Error duplicating string in table file routine.\n" );
 	strncpy(s, strToUpper(infile),strlen(infile));
   	inputbufSize = fileSize (fp);
    printf("tablesize is %d\n", inputbufSize);
 	// indata = malloc ( bufSize );
      if ( !fread ( Translate, inputbufSize , 1, fp ))
      myshutdown ( EF_FILE_READ, "Error reading from input file.\n" );
      
	fclose(fp);
	return TRUE;
 } 

int outputfile(char *outfile) 
{
  FILE *fp;
  if (isBmp) outputbufSize = 2 * res * res ;
  if ( !( fp = fopen ( outfile, WRITE_MODE )))      
  	myshutdown ( EF_FILE_WRITE, "Error opening output file.\n" );  
; /* no such file */
 	// data = malloc ( bufSize );
      if ( !fwrite ( outdata, outputbufSize, 1, fp ))
      myshutdown ( EF_FILE_WRITE, "Error writing to output file.\n" );  
	fclose(fp);
	printf("%d bytes written\n",outputbufSize);
	return TRUE;
 } 



/************************ M A I N  ********************************/
/******************************************************************/
int main(int argc, char *argv[]) {
int validinput = FALSE;
int x ;
/* set defaults */
 inpath = "materialmap.raw";
 outpath = "HeightmapPrimary.mat";
 tabpath = "conversiontable.hex";
 
  parseArgs ( argc, argv );
	printf("%s is the inpath\n",inpath);
	printf("%s is the outpath\n",outpath);
	validinput = isFile(tabpath);
	if (!validinput) printf("didn't find conversiontable.hex,using built-in\n" );
	 else readintable(tabpath);
	validinput = isFile(inpath);
	printf("found %d file to open\n",validinput );
	if (!validinput) return 0;
	validinput = readinfile(inpath);
	for (x=0; x<12 ; x++) printf("%d ",indata[x]);
	outputbufSize = process();
	printf("\n");
	for (x=0; x<16 ; x++) printf("%d ",outdata[x]);
	outputfile(outpath);	
	return 0;
}

void
parseArgs ( int argc, char *argv[] )
{
  char argType;
  int i = 0;
  char *name;

  if (( name = strrchr ( argv[0], PATHSEP )) == NULL ) name = argv[0];
  else name++;

  while ( ++i < argc ) {

    if (( strlen (argv[i]) == 2 ) && ( argv[i][0] == COMSEP )) {
      argType = argv[i][1];
      switch (argType)
	{
	case 'V':
	  printf ( "%s Version %d.%d-%s\n",
		   name,
		   MAJOR_VERSION,
		   MINOR_VERSION,
		   SUB_VERSION );
	  exit (0);
	case 'D' : 	outputbufSize = testfill(524288); 
				outputfile(outpath);
				exit(0);
	case 'h':
	case '?':
	  /* print usage */
	  usage ( name );
	  break;
	default:
	  printf ( "Unknown parameter: %c\n", argType );
	  exit (0);
	 }
    } else {
      if  ( strlen (argv[i]) > 2 )  {
		inpath = dupString(argv[i]);

      } else if ( outpath == NULL ) {
	  outpath = dupString(argv[i]);
	 // data.outpath = "output";

      } else {
	printf ( "What do I do with this: %s ? \n", argv[i] );
	exit (0);
      }
    }
  }
} /* end parseArgs */




void
usage ( const char *name )
{
  printf ( "\nUsage: %s [options] <source file> <destination path>\n\n", name );
  printf ( "%s reads a BF1942 Materialmap and converts it to a BF2 compatible materialmap.\n", name );
  printf ( " communityware mschoeldgen 2007. Options are: \n" );
  printf ( "-h(elp)  or ? :outputs this help \n" );
  printf ( "-D(ump) outputs a test file \n" );
  printf ( "-V(ersion) outputs version data  \n" );
  exit(0);
} /* end usage */
